library WarpTower initializer preload requires AIDS, Damage, TimerUtils, LineSegment, UnitHistory, UnitCoordinateHistory, UnitDistanceHistory, UnitFacingHistory, UnitMoveOrderHistory, optional SmoothUnitMod
	globals
	//===================================================================
	// Rawcodes and boring data.
	
		//Units
		private constant integer TOWER_UNIT_ID = 'n000' //Rawcode of "Warp Tower"
		private constant integer GROUND_DUMMY_ID = 'h002' //Rawcode of "Warp Tower Dummy (Ground)"
		private constant integer AIR_DUMMY_ID = 'h001' //Rawcode of "Warp Tower Dummy (Flying)"
		
		//Abilities
		private constant integer WALL_ABILITY = 'A003' //Rawcode of "Warp Wall (Warp Tower)"
		private constant integer BUFF_AURA_ABILITY = 'A002' //Rawcode of "Recently Warped Buff (Warp Tower)"
		private constant integer ATTACK_FX_ABILITY = 'A006' //Rawcode of "Directional Warp Attack FX (Warp Tower)"
		private constant string ATTACK_FX_ORDER = "shadowstrike" //Order string for "Directional Attack FX (Warp Tower)"
		private constant integer WARP_FX_ABILITY = 'A007' //Rawcode of "Warp End FX (Warp Tower)"
		private constant string WARP_FX_ORDER = "shadowstrike" //Order string for "Warp End FX (Warp Tower)"
		private constant integer WALL_FX_ABILITY = 'A004' //Rawcode of "Warp Wall FX (Warp Tower)"
		private constant string WALL_FX_ORDER = "absorb" //Order string for "Warp Wall FX (Warp Tower)"
		
		//Buffs
		private constant integer BUFF_AURA_BUFF = 'B000' //Rawcode of Aura Buff (Warp Tower)
		private constant integer NULL_FX_BUFF = 'B001' //Rawcode of Null FX Buff (Warp Tower)
		
		//Effects
		private constant boolean ATTACK_EFFECT = true //Set to false if you don't want any additional effects when the tower attacks
		private constant boolean FADE_EFFECT = true //Set to false if you don't want to fade units out/in when warping, requires SmoothUnitMod
		private constant real FADE_DURATION = 0.5 //The time taken to fade the unit to transparent when warping
		private constant string WARP_TRAIL_FX = "war3mapImported\\WarpTowerTrail.mdx" //An effect used to trace out the target's path while warping
		private constant real GROUND_OFFSET = 60. //The amount to assume a ground unit's average attachment height should be
		
		//Unittype used to keep multiple warp towers from attacking the same unit.  The tower has this type required to attack.
		private constant unittype ATTACKS_ALLOWED = UNIT_TYPE_SAPPER
		
		//How long to wait to re-enable the target's attackability if it is unable to be warped.
		private constant real ATTACK_DISABLE_TIMEOUT = 1.
		
		//The distance above the terrain at the source of the tower's attacks.
		private constant real TOWER_ATTACK_HEIGHT = 250.
		
	//===================================================================
	// Configurables.
		
		//The base minimum amount of time to revert a target.
		private constant real TIME_REVERTED = 5.
		
		//The base minimum distance to revert a target.
		private constant real DISTANCE_REVERTED = 1000.
		
		//How much less to revert a target if it's magic immune.
		private constant real IMMUNE_FACTOR = 0.33
		
		//How long the "recently warped" buff will last.
		private constant real BUFF_DURATION = 30.
		
		//Maximum stack level for the "recently warped buff; increases when a target is warped.  At or above this level, it cannot be warped until a buff wears off.
		private constant integer MAX_BUFF_STACK = 3 //Should match the # of levels of the BUFF_AURA_ABILITY
	
		//Length of a warp wall.
		private constant real WALL_LENGTH = 600.
		
		//Width of a warp wall.
		private constant real WALL_COLLISION = 50.
		
		//Period at which to check for units crossing a warp wall.
		private constant real WALL_PERIOD = 0.15
		
		//Amount of FX to create along the wall per period.
		private constant integer WALL_EFFECTS_PER_PERIOD = 4
		
		//Mana regeneration per second for a tower that is not attacking or casting Warp Wall.
		private constant real MANA_REGEN_IDLE = 1.25
		
		//Mana regeneration per second for a tower that is attacking or casting Warp Wall.
		private constant real MANA_REGEN_BUSY = 0.25
		
		//Period at which to restore mana.  Mana restored per period is regen*period.
		private constant real MANA_REGEN_PERIOD = 0.25
		
		//Time after an attack that the tower will still be considered "attacking".  Should match the tower's attack rate.
		private constant real ATTACK_REGEN_DELAY = 4.
	endglobals
	
	//Function that multiplies how far back in time to warp a unit.
	private keyword WarpStatus
	private function ScaleFactor takes unit target returns real
		local real result = Pow(0.6, target:WarpStatus.recentWarps)
		if IsUnitType(target, UNIT_TYPE_MAGIC_IMMUNE) then
			return result*IMMUNE_FACTOR
		endif
		return result
	endfunction
	
	//Function that determines whether a unit is a valid target in a warp wall.
	private function TargetAllowed takes unit target, unit source returns boolean
		return UnitAlive(target) and IsUnitType(target, UNIT_TYPE_STRUCTURE) == false and IsUnitEnemy(target, GetOwningPlayer(source))
	endfunction
	
	//===================================================================
	// System code begins here.  Don't edit unless you're confident!
	//===================================================================
	
	native UnitAlive takes unit id returns boolean
	
	//===================================================================
	//Leak prevention globals.
	globals
		private group GROUP = CreateGroup()
		private location Z_LOC = Location(0, 0)
	endglobals
	
	private function GetCoordinateZ takes real x, real y returns real
		call MoveLocation(Z_LOC, x, y)
		return GetLocationZ(Z_LOC)
	endfunction
	
	//===================================================================
	//Dummy unit recycling stack.
	globals
		private unit array GROUND_DUMMY
		private integer GROUND_DUMMY_COUNT = 0
		private integer MAX_GROUND_DUMMIES = 25
		private unit array AIR_DUMMY
		private integer AIR_DUMMY_COUNT = 0
		private integer MAX_AIR_DUMMIES = 25
		private unit CASTER_DUMMY
	endglobals
	
	private function GetDummy takes boolean air returns unit
		if air then
			if AIR_DUMMY_COUNT == 0 then
				set AIR_DUMMY[0] = CreateUnit(Player(PLAYER_NEUTRAL_PASSIVE), AIR_DUMMY_ID, 0, 0, 0)
				call UnitRemoveAbility(AIR_DUMMY[0], 'Amov')
				call SetUnitTurnSpeed(AIR_DUMMY[0], 1.)
			else
				set AIR_DUMMY[0] = AIR_DUMMY[AIR_DUMMY_COUNT]
				set AIR_DUMMY_COUNT = AIR_DUMMY_COUNT - 1
			endif
			return AIR_DUMMY[0]
		else
			if GROUND_DUMMY_COUNT == 0 then
				set GROUND_DUMMY[0] = CreateUnit(Player(PLAYER_NEUTRAL_PASSIVE), GROUND_DUMMY_ID, 0, 0, 0)
				call UnitAddAbility(GROUND_DUMMY[0], 'Amrf')
				call UnitRemoveAbility(GROUND_DUMMY[0], 'Amrf')
				call UnitRemoveAbility(GROUND_DUMMY[0], 'Amov')
				call SetUnitTurnSpeed(GROUND_DUMMY[0], 1.)
			else
				set GROUND_DUMMY[0] = GROUND_DUMMY[GROUND_DUMMY_COUNT]
				set GROUND_DUMMY_COUNT = GROUND_DUMMY_COUNT - 1
			endif
			return GROUND_DUMMY[0]
		endif
	endfunction
	
	private function ReleaseDummy takes unit u returns nothing
		call TriggerSleepAction(0.5)
		if GetUnitTypeId(u) == AIR_DUMMY_ID then
			if AIR_DUMMY_COUNT < MAX_AIR_DUMMIES then
				set AIR_DUMMY_COUNT = AIR_DUMMY_COUNT + 1
				set AIR_DUMMY[AIR_DUMMY_COUNT] = u
			else
				call KillUnit(u)
			endif
		elseif GetUnitTypeId(u) == GROUND_DUMMY_ID then
			if GROUND_DUMMY_COUNT < MAX_GROUND_DUMMIES then
				set GROUND_DUMMY_COUNT = GROUND_DUMMY_COUNT + 1
				set GROUND_DUMMY[GROUND_DUMMY_COUNT] = u
			else
				call KillUnit(u)
			endif
		endif
	endfunction
	
	//===================================================================
	// Tower code begins here.  Don't edit unless you're confident!
	//===================================================================
	//Function that determines whether the damage source is this tower.
	private function Ident takes unit u returns boolean
		return GetUnitTypeId(u) == TOWER_UNIT_ID
	endfunction
	
	//===================================================================
	//Function that gets which unit history step is the goal for reversion.
	private function GetReversionStep takes unit target returns integer
		local integer goalStep = target:UnitHistory.toSteps(TIME_REVERTED*ScaleFactor(target))
		local real goalDistance = DISTANCE_REVERTED*ScaleFactor(target)
		local integer step = goalStep
		local real startDistance = target:DistanceHistory.step(0).get()
		local real distance = startDistance - target:DistanceHistory.step(step).get()
		loop
			exitwhen distance >= goalDistance
			set step = step + 1
			exitwhen not target:DistanceHistory.check(step)
			set distance = startDistance - target:DistanceHistory.step(step).get()
		endloop
		return step
	endfunction
	
	//===================================================================
	// Struct regarding the warp status of a unit.
	private struct WarpStatus extends array
		boolean warping
		integer recentWarps
		boolean targeted
		timer targetingTimer
		
		method incrementRecentWarps takes nothing returns nothing
			set recentWarps = recentWarps + 1
			call UnitAddAbility(.unit, BUFF_AURA_ABILITY)
			call SetUnitAbilityLevel(.unit, BUFF_AURA_ABILITY, recentWarps)
		endmethod
		
		method decrementRecentWarps takes nothing returns nothing
			set recentWarps = recentWarps - 1
			if recentWarps == 0 then
				call UnitRemoveAbility(.unit, BUFF_AURA_ABILITY)
				call UnitRemoveAbility(.unit, BUFF_AURA_BUFF)
			else
				call SetUnitAbilityLevel(.unit, BUFF_AURA_ABILITY, recentWarps)
			endif
		endmethod
		
		private static method targetingCallback takes nothing returns nothing
			local thistype this = thistype(GetTimerData(GetExpiredTimer()))
			call ReleaseTimer(targetingTimer)
			set targetingTimer = null
			set targeted = false
			call this.allowAttacks()
		endmethod
		
		method targeting takes nothing returns nothing
			if not targeted then
				set targeted = true
				set targetingTimer = NewTimer()
				call SetTimerData(targetingTimer, this)
				call TimerStart(targetingTimer, ATTACK_DISABLE_TIMEOUT, false, function thistype.targetingCallback)
			endif
		endmethod
		
		method allowAttacks takes nothing returns nothing
			if not warping and not targeted and recentWarps < MAX_BUFF_STACK then
				call UnitAddType(.unit, ATTACKS_ALLOWED)
			endif
		endmethod
		
		method preventAttacks takes nothing returns nothing
				call UnitRemoveType(.unit, ATTACKS_ALLOWED)
			if not targeted then
				set targeted = true
				set targetingTimer = NewTimer()
				call SetTimerData(targetingTimer, this)
				call TimerStart(targetingTimer, ATTACK_DISABLE_TIMEOUT, false, function thistype.targetingCallback)
			endif
		endmethod
		
		//! runtextmacro AIDS()
		private method AIDS_onCreate takes nothing returns nothing
			set warping = false
			set recentWarps = 0
			set targeted = false
			call .allowAttacks()
		endmethod
		private method AIDS_onDestroy takes nothing returns nothing
			if targetingTimer != null then
				call ReleaseTimer(targetingTimer)
				set targetingTimer = null
			endif
		endmethod
	endstruct
	
	//===================================================================
	// Struct controlling the effect of a unit being warped.
	private struct WarpEffect
		private unit whichUnit
		private unit caster
		private integer goalStep
		private integer thisStep = 0
		private unit dummy
		private effect trail
		
		private static method endBuff takes nothing returns nothing
			local thistype this = thistype(GetTimerData(GetExpiredTimer()))
			if whichUnit:WarpStatus != 0 then
				call whichUnit:WarpStatus.decrementRecentWarps()
				if whichUnit:WarpStatus.recentWarps == MAX_BUFF_STACK - 1 then
					call whichUnit:WarpStatus.allowAttacks()
				endif
			endif
			call ReleaseTimer(GetExpiredTimer())
			call this.destroy()
		endmethod
		
		static method warpFX takes unit target returns nothing
			call UnitAddAbility(CASTER_DUMMY, WARP_FX_ABILITY)
			call IssueTargetOrder(CASTER_DUMMY, WARP_FX_ORDER, target)
			call UnitRemoveAbility(CASTER_DUMMY, WARP_FX_ABILITY)
		endmethod
		
		static method periodic takes nothing returns nothing
			local thistype this = thistype(GetTimerData(GetExpiredTimer()))
			local WarpStatus status = whichUnit:WarpStatus
			if thisStep < goalStep then
				call SetUnitX(.dummy, whichUnit:XHistory.step(thisStep).get())
				call SetUnitY(.dummy, whichUnit:YHistory.step(thisStep).get())
				set thisStep = thisStep + 1
			else
				call SetUnitPosition(whichUnit, whichUnit:XHistory.step(thisStep).get(), whichUnit:YHistory.step(thisStep).get())
				call whichUnit:UnitHistory.unsuspend()
				call whichUnit:UnitHistory.unlock()
				call whichUnit:XHistory.step(thisStep).trim()
				call whichUnit:YHistory.step(thisStep).trim()
				call whichUnit:DistanceHistory.step(thisStep).trim()
				call whichUnit:FacingHistory.step(thisStep).restore().trim()
				call whichUnit:MoveOrderHistory.step(thisStep).restore().trim()
				set status.warping = false
				call status.allowAttacks()
				
				static if LIBRARY_SmoothUnitMod and FADE_EFFECT then
					call FadeUnitRGB(whichUnit, FADE_DURATION, 255., 127., 0., 255.)
				endif
				call thistype.warpFX(whichUnit)
				
				call DestroyEffect(trail)
				call ReleaseDummy.execute(dummy)
				
				call ReleaseTimer(GetExpiredTimer())
			endif
		endmethod
		
		static method start takes unit target, unit source returns nothing
			local thistype this
			local timer t
			local WarpStatus status = target:WarpStatus
			
			set this = thistype.create()
			set whichUnit = target
			set caster = source
			set status.warping = true
			call status.preventAttacks()
			call status.incrementRecentWarps()
			static if LIBRARY_SmoothUnitMod and FADE_EFFECT then
				call FadeUnitRGB(whichUnit, FADE_DURATION, -255., -127., 0., -255.)
			endif
			
			set t = NewTimer()
			call SetTimerData(t, this)
			call TimerStart(t, BUFF_DURATION, false, function thistype.endBuff)
			
			set goalStep = GetReversionStep(target)
			call whichUnit:FacingHistory.step(goalStep).restore()
			call whichUnit:UnitHistory.suspend()
			call whichUnit:UnitHistory.lock()
			
			if IsUnitType(whichUnit, UNIT_TYPE_FLYING) == true then
				set dummy = GetDummy(true)
				call SetUnitFlyHeight(dummy, GetUnitFlyHeight(whichUnit), 0.)
			else
				set dummy = GetDummy(false)
				call SetUnitFlyHeight(dummy, GetUnitFlyHeight(whichUnit) + GROUND_OFFSET, 0.)
			endif
			call SetUnitX(dummy, GetUnitX(whichUnit))
			call SetUnitY(dummy, GetUnitY(whichUnit))
			set trail = AddSpecialEffectTarget(WARP_TRAIL_FX, .dummy, "origin")
			set t = NewTimer()
			call SetTimerData(t, this)
			call TimerStart(t, FADE_DURATION/I2R(goalStep), true, function thistype.periodic)
		endmethod
	endstruct
	
	//===================================================================
	// Struct controlling the warp wall ability of the tower, once cast.
	private keyword WarpTower
	
	private struct WarpWall
		private unit caster
		private real sourceX
		private real sourceY
		private real sourceZ
		private real targetX
		private real targetY
		private timer wallTimer
		private static unit wallDummy
		
		private static unit source //Global to pass into the enum
		
		private static method enum takes nothing returns nothing
			local unit target = GetEnumUnit()
			if target:WarpStatus.recentWarps < MAX_BUFF_STACK and not target:WarpStatus.warping then
				call WarpEffect.warpFX(target)
				call WarpEffect.start(target, thistype.source)
			endif
			set target = null
		endmethod
		
		private static method filter takes nothing returns boolean
			return TargetAllowed(GetFilterUnit(), thistype.source)
		endmethod
		
		private static method periodic takes nothing returns nothing
			local thistype this = thistype(GetTimerData(GetExpiredTimer()))
			local real factor = 0
			local real x = 0
			local real y = 0
			local integer index = WALL_EFFECTS_PER_PERIOD
			call UnitRemoveAbility(wallDummy, 'Aloc')
			call UnitAddAbility(wallDummy, WALL_FX_ABILITY)
			call SetUnitState(wallDummy, UNIT_STATE_MANA, 1.)
			loop
				exitwhen index == 0
				set index = index - 1
				
				set factor = GetRandomReal(0.1, 1.)
				set x = factor*(targetX - sourceX) + sourceX
				set y = factor*(targetY - sourceY) + sourceY
				
				call SetUnitX(wallDummy, x)
				call SetUnitY(wallDummy, y)
				
				call SetUnitFlyHeight(wallDummy, 0., 0.)
				call IssueTargetOrder(wallDummy, WALL_FX_ORDER, wallDummy)
				
				call SetUnitFlyHeight(wallDummy, sourceZ - GetCoordinateZ(x, y) + TOWER_ATTACK_HEIGHT, 0.)
				call IssueTargetOrder(wallDummy, WALL_FX_ORDER, wallDummy)
			endloop
			call UnitRemoveAbility(wallDummy, WALL_FX_ABILITY)
			static if ATTACK_EFFECT then
				call WarpTower.attackFX(wallDummy, .caster)
			endif
			call UnitAddAbility(wallDummy, 'Aloc')
			set source = caster
			call GroupEnumUnitsInRangeOfSegment(GROUP, sourceX, sourceY, targetX, targetY, WALL_COLLISION, Filter(function thistype.filter))
			call ForGroup(GROUP, function thistype.enum)
			call GroupClear(GROUP)
		endmethod
		
		static method create takes unit u, real X, real Y, real Z returns thistype
			local thistype this = thistype.allocate()
			local real tx = 0
			local real ty = 0 
			local real factor = 0
			set caster = u
			set sourceX = GetUnitX(u)
			set sourceY = GetUnitY(u)
			set sourceZ = Z
			set tx = X - sourceX
			set ty = Y - sourceY
			set factor = WALL_LENGTH/SquareRoot(tx*tx + ty*ty)
			set targetX = sourceX + tx*factor
			set targetY = sourceY + ty*factor
			set wallTimer = NewTimer()
			call SetTimerData(wallTimer, this)
			call TimerStart(wallTimer, WALL_PERIOD, true, function thistype.periodic)
			return this
		endmethod
		
		private method onDestroy takes nothing returns nothing
			call ReleaseTimer(wallTimer)
		endmethod
		
		private static method onInit takes nothing returns nothing
			set wallDummy = GetDummy(false)
		endmethod
	endstruct
	
	//===================================================================
	// Struct with information and methods pertaining to the tower itself.
	private struct WarpTower extends array
		private WarpWall wall
		private real terrainZ
		private real regenDelay
		
		private static unit array towerStack
		private static integer towerCount = 0
		private integer thisStackIndex
		
		private static method castWall takes nothing returns boolean
			if GetSpellAbilityId() == WALL_ABILITY then
				set GetTriggerUnit():thistype.wall = WarpWall.create(GetTriggerUnit(), GetSpellTargetX(), GetSpellTargetY(), GetTriggerUnit():thistype.terrainZ)
			endif
			return false
		endmethod
		
		private static method stopWall takes nothing returns boolean
			if GetSpellAbilityId() == WALL_ABILITY then
				call GetTriggerUnit():thistype.wall.destroy()
			endif
			return false
		endmethod
		
		static method attackFX takes unit target, unit source returns nothing
			call SetUnitX(CASTER_DUMMY, GetUnitX(source))
			call SetUnitY(CASTER_DUMMY, GetUnitY(source))
			call SetUnitFlyHeight(CASTER_DUMMY, TOWER_ATTACK_HEIGHT, 0.)
			call UnitAddAbility(CASTER_DUMMY, ATTACK_FX_ABILITY)
			call IssueTargetOrder(CASTER_DUMMY, ATTACK_FX_ORDER, target)
			call UnitRemoveAbility(CASTER_DUMMY, ATTACK_FX_ABILITY)
		endmethod
		
		private static method onAttack takes nothing returns boolean
			if Ident(GetAttacker()) then
				set GetAttacker():WarpTower.regenDelay = ATTACK_REGEN_DELAY
				call GetTriggerUnit():WarpStatus.preventAttacks()
				call GetTriggerUnit():WarpStatus.targeting()
				static if ATTACK_EFFECT then
					call attackFX(GetTriggerUnit(), GetAttacker())
				endif
			endif
			return false
		endmethod
		
		private static method attackHit takes nothing returns boolean
			local unit target = GetTriggerUnit()
			if Ident(GetEventDamageSource()) and Damage_IsAttack() then
				if target:WarpStatus.recentWarps < MAX_BUFF_STACK and not target:WarpStatus.warping then
					call WarpEffect.start(GetTriggerUnit(), GetEventDamageSource())
				endif
			endif
			call UnitRemoveAbility(GetTriggerUnit(), NULL_FX_BUFF)
			return false
		endmethod
		
		private static method manaRegen takes nothing returns nothing
			local integer index = towerCount
			loop
				exitwhen index == 0
				if UnitAlive(towerStack[index]) and GetUnitCurrentOrder(towerStack[index]) == 0 and towerStack[index]:WarpTower.regenDelay == 0 then
					call SetUnitState(towerStack[index], UNIT_STATE_MANA, GetUnitState(towerStack[index], UNIT_STATE_MANA) + MANA_REGEN_IDLE*MANA_REGEN_PERIOD)
				else
					
					call SetUnitState(towerStack[index], UNIT_STATE_MANA, GetUnitState(towerStack[index], UNIT_STATE_MANA) + MANA_REGEN_BUSY*MANA_REGEN_PERIOD)
				endif
				if towerStack[index]:WarpTower.regenDelay > 0 then
					set towerStack[index]:WarpTower.regenDelay = towerStack[index]:WarpTower.regenDelay - MANA_REGEN_PERIOD
				endif
				set index = index - 1
			endloop
		endmethod
		
		//! runtextmacro AIDS()
		private static method AIDS_filter takes unit u returns boolean
			return Ident(u)
		endmethod
		
		private method AIDS_onCreate takes nothing returns nothing
			set terrainZ = GetCoordinateZ(GetUnitX(.unit), GetUnitY(.unit))
			set regenDelay = 0
			set towerCount = towerCount + 1
			set towerStack[towerCount] = .unit
			set thisStackIndex = towerCount
		endmethod
		
		private method AIDS_onDestroy takes nothing returns nothing
			set towerStack[thisStackIndex] = towerStack[towerCount]
			set towerStack[towerCount]:WarpTower.thisStackIndex = thisStackIndex
			set towerCount = towerCount - 1
		endmethod
		
		private static method AIDS_onInit takes nothing returns nothing
			local trigger t = CreateTrigger()
			
			call Damage_RegisterZeroEvent(t)
			call TriggerAddCondition(t, Condition(function thistype.attackHit))
			set t = CreateTrigger()
			call TriggerRegisterAnyUnitEventBJ(t, EVENT_PLAYER_UNIT_ATTACKED)
			call TriggerAddCondition(t, Condition(function thistype.onAttack))
			set t = CreateTrigger()
			call TriggerRegisterAnyUnitEventBJ(t, EVENT_PLAYER_UNIT_SPELL_EFFECT)
			call TriggerAddCondition(t, Condition(function thistype.castWall))
			set t = CreateTrigger()
			call TriggerRegisterAnyUnitEventBJ(t, EVENT_PLAYER_UNIT_SPELL_ENDCAST)
			call TriggerAddCondition(t, Condition(function thistype.stopWall))
			set t = null
			call TimerStart(CreateTimer(), MANA_REGEN_PERIOD, true, function thistype.manaRegen)
		endmethod
	endstruct
	
	private function preload takes nothing returns nothing
		set CASTER_DUMMY = GetDummy(false)
		call UnitAddAbility(CASTER_DUMMY, WALL_ABILITY)
		call UnitRemoveAbility(CASTER_DUMMY, WALL_ABILITY)
		call UnitAddAbility(CASTER_DUMMY, BUFF_AURA_ABILITY)
		call UnitRemoveAbility(CASTER_DUMMY, BUFF_AURA_ABILITY)
		call UnitAddAbility(CASTER_DUMMY, ATTACK_FX_ABILITY)
		call UnitRemoveAbility(CASTER_DUMMY, ATTACK_FX_ABILITY)
		call UnitAddAbility(CASTER_DUMMY, WALL_FX_ABILITY)
		call UnitRemoveAbility(CASTER_DUMMY, WALL_FX_ABILITY)
	endfunction
endlibrary
